@@ -3,7 +3,6 @@ package daemon
 import (
 	"bufio"
 	"bytes"
-	"encoding/json"
 	"fmt"
 	"io/ioutil"
 	"os"
@@ -17,20 +16,24 @@ import (
 	"syscall"
 	"time"
 
+	"github.com/clarketm/json"
+	ign "github.com/coreos/ignition/config/v2_2"
 	igntypes "github.com/coreos/ignition/config/v2_2/types"
 	"github.com/golang/glog"
 	"github.com/google/renameio"
-	drain "github.com/openshift/cluster-api/pkg/drain"
-	mcfgv1 "github.com/openshift/machine-config-operator/pkg/apis/machineconfiguration.openshift.io/v1"
-	ctrlcommon "github.com/openshift/machine-config-operator/pkg/controller/common"
-	"github.com/openshift/machine-config-operator/pkg/daemon/constants"
-	pivotutils "github.com/openshift/machine-config-operator/pkg/daemon/pivot/utils"
 	errors "github.com/pkg/errors"
 	"github.com/vincent-petithory/dataurl"
 	corev1 "k8s.io/api/core/v1"
 	metav1 "k8s.io/apimachinery/pkg/apis/meta/v1"
+	"k8s.io/apimachinery/pkg/runtime"
 	"k8s.io/apimachinery/pkg/util/uuid"
 	"k8s.io/apimachinery/pkg/util/wait"
+	"k8s.io/kubectl/pkg/drain"
+
+	mcfgv1 "github.com/openshift/machine-config-operator/pkg/apis/machineconfiguration.openshift.io/v1"
+	ctrlcommon "github.com/openshift/machine-config-operator/pkg/controller/common"
+	"github.com/openshift/machine-config-operator/pkg/daemon/constants"
+	pivotutils "github.com/openshift/machine-config-operator/pkg/daemon/pivot/utils"
 )
 
 const (
@@ -44,12 +47,25 @@ const (
 	coreUserSSHPath = "/home/core/.ssh/"
 	// fipsFile is the file to check if FIPS is enabled
 	fipsFile = "/proc/sys/crypto/fips_enabled"
-	// Traditional kernel
-	defaultKernelType = "default"
-	// Realtime kernel
-	realtimeKernelType = "realtime"
 )
 
+func installedRTKernelRpmsOnHost() ([]string, error) {
+	var err error
+	var rtKernelRpms = []string{}
+	var bootedDeployment *RpmOstreeDeployment
+	client := NewNodeUpdaterClient()
+	if bootedDeployment, err = client.GetBootedDeployment(); err != nil {
+		return rtKernelRpms, err
+	}
+
+	for _, localPkg := range bootedDeployment.RequestedLocalPkgs {
+		if strings.HasPrefix(localPkg, "kernel-rt-") {
+			rtKernelRpms = append(rtKernelRpms, localPkg)
+		}
+	}
+	return rtKernelRpms, nil
+}
+
 func writeFileAtomicallyWithDefaults(fpath string, b []byte) error {
 	return writeFileAtomically(fpath, b, defaultDirectoryPermissions, defaultFilePermissions, -1, -1)
 }
@@ -93,16 +109,6 @@ func getNodeRef(node *corev1.Node) *corev1.ObjectReference {
 	}
 }
 
-type drainLogger struct{}
-
-func (dl *drainLogger) Logf(format string, v ...interface{}) {
-	glog.Infof(format, v...)
-}
-
-func (dl *drainLogger) Log(v ...interface{}) {
-	glog.Info(v...)
-}
-
 // Remove pending deployment on OSTree based system
 func removePendingDeployment() error {
 	_, err := runGetOut("rpm-ostree", "cleanup", "-p")
@@ -182,13 +188,13 @@ func (dn *Daemon) drain() error {
 	}
 	var lastErr error
 	if err := wait.ExponentialBackoff(backoff, func() (bool, error) {
-		err := drain.Drain(dn.kubeClient, []*corev1.Node{dn.node}, &drain.DrainOptions{
-			DeleteLocalData:    true,
-			Force:              true,
-			GracePeriodSeconds: -1,
-			IgnoreDaemonsets:   true,
-			Logger:             &drainLogger{},
-		})
+		err := drain.RunCordonOrUncordon(dn.drainer, dn.node, true)
+		if err != nil {
+			lastErr = err
+			glog.Infof("Cordon failed with: %v, retrying", err)
+			return false, nil
+		}
+		err = drain.RunNodeDrain(dn.drainer, dn.node.Name)
 		if err == nil {
 			return true, nil
 		}
@@ -200,9 +206,11 @@ func (dn *Daemon) drain() error {
 		if err == wait.ErrWaitTimeout {
 			failMsg := fmt.Sprintf("%d tries: %v", backoff.Steps, lastErr)
 			MCDDrainErr.WithLabelValues(failTime, failMsg).SetToCurrentTime()
+			dn.recorder.Eventf(getNodeRef(dn.node), corev1.EventTypeWarning, "FailedToDrain", failMsg)
 			return errors.Wrapf(lastErr, "failed to drain node (%d tries): %v", backoff.Steps, err)
 		}
 		MCDDrainErr.WithLabelValues(failTime, err.Error()).SetToCurrentTime()
+		dn.recorder.Eventf(getNodeRef(dn.node), corev1.EventTypeWarning, "FailedToDrain", err.Error())
 		return errors.Wrap(err, "failed to drain node")
 	}
 
@@ -221,51 +229,59 @@ func canonicalizeEmptyMC(config *mcfgv1.MachineConfig) *mcfgv1.MachineConfig {
 	if config != nil {
 		return config
 	}
+	newIgnCfg := ctrlcommon.NewIgnConfig()
+	rawNewIgnCfg, err := json.Marshal(newIgnCfg)
+	if err != nil {
+		// This should never happen
+		panic(err)
+	}
 	return &mcfgv1.MachineConfig{
 		ObjectMeta: metav1.ObjectMeta{Name: "mco-empty-mc"},
 		Spec: mcfgv1.MachineConfigSpec{
-			Config: ctrlcommon.NewIgnConfig(),
+			Config: runtime.RawExtension{
+				Raw: rawNewIgnCfg,
+			},
 		},
 	}
 }
 
-// Returns true if updated packages are available
-func rtKernelUpdateAvailable(rpms []os.FileInfo, rtKernelPkg []string) (bool, error) {
-	for _, pkg := range rtKernelPkg {
-		var out []byte
-		var err error
+// Returns true if updated packages or new RT kernel related packages are available
+func rtKernelUpdateAvailable(updateRpms []os.FileInfo, installedRTKernelRpms []string) bool {
+	// if list of RT kernel packages in update is more than installed list, then we have additional packages to install
+	if len(updateRpms) > len(installedRTKernelRpms) {
+		return true
+	}
+	for _, pkg := range installedRTKernelRpms {
 		found := false
-
-		if out, err = exec.Command("rpm", "-q", pkg).Output(); err != nil {
-			wrappedErr := fmt.Errorf("Failed to run rpm -q : %v", err)
-			return false, wrappedErr
-		}
-		searchRpm := strings.TrimSpace(string(out)) + ".rpm"
-		for _, rpm := range rpms {
+		searchRpm := pkg + ".rpm"
+		for _, rpm := range updateRpms {
 			if rpm.Name() == searchRpm {
 				found = true
 				break
 			}
 		}
 		if !found {
-			return true, nil
+			return true
 		}
 	}
 
-	return false, nil
+	return false
 }
 
-func (dn *Daemon) compareMachineConfig(oldConfig, newConfig *mcfgv1.MachineConfig) bool {
-	var diff *MachineConfigDiff
+// return true if the MachineConfigDiff is not empty
+func (dn *Daemon) compareMachineConfig(oldConfig, newConfig *mcfgv1.MachineConfig) (bool, error) {
 	oldConfig = canonicalizeEmptyMC(oldConfig)
 	oldConfigName := oldConfig.GetName()
 	newConfigName := newConfig.GetName()
-	diff = NewMachineConfigDiff(oldConfig, newConfig)
-	if diff.IsEmpty() {
-		glog.Infof("No changes from %s to %s", oldConfigName, newConfigName)
-		return false
+	mcDiff, err := NewMachineConfigDiff(oldConfig, newConfig)
+	if err != nil {
+		return true, errors.Wrapf(err, "error creating MachineConfigDiff for comparison")
 	}
-	return true
+	if mcDiff.IsEmpty() {
+		glog.Infof("No changes from %s to %s", oldConfigName, newConfigName)
+		return false, nil
+	}
+	return true, nil
 }
 
 // update the node to the provided node configuration.
@@ -332,13 +348,22 @@ func (dn *Daemon) update(oldConfig, newConfig *mcfgv1.MachineConfig) (retErr err
 		}
 	}()
 
-	if err := dn.updateSSHKeys(newConfig.Spec.Config.Passwd.Users); err != nil {
+	oldIgnConfig, report, err := ign.Parse(oldConfig.Spec.Config.Raw)
+	if err != nil {
+		return fmt.Errorf("parsing old Ignition config failed with error: %v\nReport: %v", err, report)
+	}
+	newIgnConfig, report, err := ign.Parse(newConfig.Spec.Config.Raw)
+	if err != nil {
+		return fmt.Errorf("parsing new Ignition config failed with error: %v\nReport: %v", err, report)
+	}
+
+	if err := dn.updateSSHKeys(newIgnConfig.Passwd.Users); err != nil {
 		return err
 	}
 
 	defer func() {
 		if retErr != nil {
-			if err := dn.updateSSHKeys(oldConfig.Spec.Config.Passwd.Users); err != nil {
+			if err := dn.updateSSHKeys(oldIgnConfig.Passwd.Users); err != nil {
 				retErr = errors.Wrapf(retErr, "error rolling back SSH keys updates %v", err)
 				return
 			}
@@ -389,16 +414,22 @@ type MachineConfigDiff struct {
 
 // canonicalizeKernelType returns a valid kernelType. We consider empty("") and default kernelType as same
 func canonicalizeKernelType(kernelType string) string {
-	if kernelType == realtimeKernelType {
-		return realtimeKernelType
+	if kernelType == ctrlcommon.KernelTypeRealtime {
+		return ctrlcommon.KernelTypeRealtime
 	}
-	return defaultKernelType
+	return ctrlcommon.KernelTypeDefault
 }
 
 // NewMachineConfigDiff compares two MachineConfig objects.
-func NewMachineConfigDiff(oldConfig, newConfig *mcfgv1.MachineConfig) *MachineConfigDiff {
-	oldIgn := oldConfig.Spec.Config
-	newIgn := newConfig.Spec.Config
+func NewMachineConfigDiff(oldConfig, newConfig *mcfgv1.MachineConfig) (*MachineConfigDiff, error) {
+	oldIgn, report, err := ign.Parse(oldConfig.Spec.Config.Raw)
+	if err != nil {
+		return nil, fmt.Errorf("parsing old Ignition config failed with error: %v\nReport: %v", err, report)
+	}
+	newIgn, report, err := ign.Parse(newConfig.Spec.Config.Raw)
+	if err != nil {
+		return nil, fmt.Errorf("parsing new Ignition config failed with error: %v\nReport: %v", err, report)
+	}
 
 	// Both nil and empty slices are of zero length,
 	// consider them as equal while comparing KernelArguments in both MachineConfigs
@@ -412,7 +443,7 @@ func NewMachineConfigDiff(oldConfig, newConfig *mcfgv1.MachineConfig) *MachineCo
 		files:      !reflect.DeepEqual(oldIgn.Storage.Files, newIgn.Storage.Files),
 		units:      !reflect.DeepEqual(oldIgn.Systemd.Units, newIgn.Systemd.Units),
 		kernelType: canonicalizeKernelType(oldConfig.Spec.KernelType) != canonicalizeKernelType(newConfig.Spec.KernelType),
-	}
+	}, nil
 }
 
 // IsEmpty returns true if the MachineConfigDiff has no changes, or
@@ -434,26 +465,22 @@ func (d *MachineConfigDiff) IsEmpty() bool {
 // directories, links, and systemd units sections of the included ignition
 // config currently.
 func Reconcilable(oldConfig, newConfig *mcfgv1.MachineConfig) (*MachineConfigDiff, error) {
-	oldIgn := oldConfig.Spec.Config
-	newIgn := newConfig.Spec.Config
+	// The parser will try to translate versions less than maxVersion to maxVersion, or output an err.
+	// The ignition output in case of success will always have maxVersion
+	oldIgn, report, err := ign.Parse(oldConfig.Spec.Config.Raw)
+	if err != nil {
+		return nil, fmt.Errorf("parsing old Ignition config failed with error: %v\nReport: %v", err, report)
+	}
+	newIgn, report, err := ign.Parse(newConfig.Spec.Config.Raw)
+	if err != nil {
+		return nil, fmt.Errorf("parsing new Ignition config failed with error: %v\nReport: %v", err, report)
+	}
 
-	// Ignition section
-	// First check if this is a generally valid Ignition Config
+	// Check if this is a generally valid Ignition Config
 	if err := ctrlcommon.ValidateIgnition(newIgn); err != nil {
 		return nil, err
 	}
 
-	// if the config versions are different, all bets are off. this probably
-	// shouldn't happen, but if it does, we can't deal with it.
-	if oldIgn.Ignition.Version != newIgn.Ignition.Version {
-		return nil, fmt.Errorf("ignition version mismatch between old and new config: old: %s new: %s",
-			oldIgn.Ignition.Version, newIgn.Ignition.Version)
-	}
-	// everything else in the ignition section doesn't matter to us, since the
-	// rest of the stuff in this section has to do with fetching remote
-	// resources, and the mcc should've fully rendered those out before the
-	// config gets here.
-
 	// Networkd section
 
 	// we don't currently configure the network in place. we can't fix it if
@@ -536,7 +563,11 @@ func Reconcilable(oldConfig, newConfig *mcfgv1.MachineConfig) (*MachineConfigDif
 
 	// we made it through all the checks. reconcile away!
 	glog.V(2).Info("Configs are reconcilable")
-	return NewMachineConfigDiff(oldConfig, newConfig), nil
+	mcDiff, err := NewMachineConfigDiff(oldConfig, newConfig)
+	if err != nil {
+		return nil, errors.Wrapf(err, "error creating MachineConfigDiff")
+	}
+	return mcDiff, nil
 }
 
 // verifyUserFields returns nil if the user Name = "core", if 1 or more SSHKeys exist for
@@ -623,8 +654,8 @@ func (dn *Daemon) updateKernelArguments(oldConfig, newConfig *mcfgv1.MachineConf
 	if len(diff) == 0 {
 		return nil
 	}
-	if dn.OperatingSystem != machineConfigDaemonOSRHCOS {
-		return fmt.Errorf("Updating kargs on non-RHCOS nodes is not supported: %v", diff)
+	if dn.OperatingSystem != machineConfigDaemonOSRHCOS && dn.OperatingSystem != machineConfigDaemonOSFCOS {
+		return fmt.Errorf("Updating kargs on non-CoreOS nodes is not supported: %v", diff)
 	}
 
 	args := append([]string{"kargs"}, diff...)
@@ -666,7 +697,7 @@ func (dn *Daemon) mountOSContainer(container string) (mnt, containerName string,
 // Right now it supports default (traditional) and realtime kernel
 func (dn *Daemon) switchKernel(oldConfig, newConfig *mcfgv1.MachineConfig) error {
 	// Do nothing if both old and new KernelType are of type default
-	if canonicalizeKernelType(oldConfig.Spec.KernelType) == defaultKernelType && canonicalizeKernelType(newConfig.Spec.KernelType) == defaultKernelType {
+	if canonicalizeKernelType(oldConfig.Spec.KernelType) == ctrlcommon.KernelTypeDefault && canonicalizeKernelType(newConfig.Spec.KernelType) == ctrlcommon.KernelTypeDefault {
 		return nil
 	}
 	// We support Kernel update only on RHCOS nodes
@@ -675,16 +706,24 @@ func (dn *Daemon) switchKernel(oldConfig, newConfig *mcfgv1.MachineConfig) error
 	}
 
 	defaultKernel := []string{"kernel", "kernel-core", "kernel-modules", "kernel-modules-extra"}
-	rtKernel := []string{"kernel-rt-core", "kernel-rt-modules", "kernel-rt-modules-extra"}
 	var args []string
 
 	dn.logSystem("Initiating switch from kernel %s to %s", canonicalizeKernelType(oldConfig.Spec.KernelType), canonicalizeKernelType(newConfig.Spec.KernelType))
 
-	if canonicalizeKernelType(oldConfig.Spec.KernelType) == realtimeKernelType && canonicalizeKernelType(newConfig.Spec.KernelType) == defaultKernelType {
+	if canonicalizeKernelType(oldConfig.Spec.KernelType) == ctrlcommon.KernelTypeRealtime && canonicalizeKernelType(newConfig.Spec.KernelType) == ctrlcommon.KernelTypeDefault {
+		var installedRTKernelRpms []string
+		var err error
 		args = []string{"override", "reset"}
 		args = append(args, defaultKernel...)
-		rtKernelUninstall := []string{"--uninstall", "kernel-rt-core", "--uninstall", "kernel-rt-modules", "--uninstall", "kernel-rt-modules-extra"}
-		args = append(args, rtKernelUninstall...)
+		if installedRTKernelRpms, err = installedRTKernelRpmsOnHost(); err != nil {
+			return fmt.Errorf("Error while fetching installed RT kernel on host %v", err)
+		}
+		if len(installedRTKernelRpms) == 0 {
+			return fmt.Errorf("No kernel-rt package installed on host")
+		}
+		for _, installedRTKernelRpm := range installedRTKernelRpms {
+			args = append(args, "--uninstall", installedRTKernelRpm)
+		}
 		dn.logSystem("Switching to kernelType=%s, invoking rpm-ostree %+q", newConfig.Spec.KernelType, args)
 		if err := exec.Command("rpm-ostree", args...).Run(); err != nil {
 			return fmt.Errorf("Failed to execute rpm-ostree %+q : %v", args, err)
@@ -724,7 +763,7 @@ func (dn *Daemon) switchKernel(oldConfig, newConfig *mcfgv1.MachineConfig) error
 		return fmt.Errorf("No kernel-rt package available in the OSContainer with URL %s", newConfig.Spec.OSImageURL)
 	}
 
-	if canonicalizeKernelType(oldConfig.Spec.KernelType) == defaultKernelType && canonicalizeKernelType(newConfig.Spec.KernelType) == realtimeKernelType {
+	if canonicalizeKernelType(oldConfig.Spec.KernelType) == ctrlcommon.KernelTypeDefault && canonicalizeKernelType(newConfig.Spec.KernelType) == ctrlcommon.KernelTypeRealtime {
 		// Switch to RT kernel
 		args = []string{"override", "remove"}
 		args = append(args, defaultKernel...)
@@ -738,23 +777,29 @@ func (dn *Daemon) switchKernel(oldConfig, newConfig *mcfgv1.MachineConfig) error
 		}
 	}
 
-	if canonicalizeKernelType(oldConfig.Spec.KernelType) == realtimeKernelType && canonicalizeKernelType(newConfig.Spec.KernelType) == realtimeKernelType {
+	if canonicalizeKernelType(oldConfig.Spec.KernelType) == ctrlcommon.KernelTypeRealtime && canonicalizeKernelType(newConfig.Spec.KernelType) == ctrlcommon.KernelTypeRealtime {
 		if oldConfig.Spec.OSImageURL != newConfig.Spec.OSImageURL {
+			var installedRTKernelRpms []string
+			var err error
 			args = []string{"uninstall"}
-			args = append(args, rtKernel...)
+			if installedRTKernelRpms, err = installedRTKernelRpmsOnHost(); err != nil {
+				return fmt.Errorf("Error while fetching installed RT kernel on host %v", err)
+			}
+			if len(installedRTKernelRpms) == 0 {
+				return fmt.Errorf("No kernel-rt package installed on host")
+			}
+			for _, installedRTKernelRpm := range installedRTKernelRpms {
+				args = append(args, installedRTKernelRpm)
+			}
 			// Perform kernel-rt package update only if updated packages are available
-			var updateAvailable bool
-			if updateAvailable, err = rtKernelUpdateAvailable(rtKernelRpms, rtKernel); err != nil {
-				return err
-			} else if !updateAvailable {
-				return nil
-			}
-			for _, rpm := range rtKernelRpms {
-				args = append(args, "--install", fmt.Sprintf("%s/%s", mnt, rpm.Name()))
-			}
-			dn.logSystem("Updating rt-kernel packages on host: %+q", args)
-			if err := exec.Command("rpm-ostree", args...).Run(); err != nil {
-				return fmt.Errorf("Failed to execute rpm-ostree %+q : %v", args, err)
+			if rtKernelUpdateAvailable(rtKernelRpms, installedRTKernelRpms) {
+				for _, rpm := range rtKernelRpms {
+					args = append(args, "--install", fmt.Sprintf("%s/%s", mnt, rpm.Name()))
+				}
+				dn.logSystem("Updating rt-kernel packages on host: %+q", args)
+				if err := exec.Command("rpm-ostree", args...).Run(); err != nil {
+					return fmt.Errorf("Failed to execute rpm-ostree %+q : %v", args, err)
+				}
 			}
 		}
 	}
@@ -781,28 +826,45 @@ func (dn *Daemon) switchKernel(oldConfig, newConfig *mcfgv1.MachineConfig) error
 // touched.
 func (dn *Daemon) updateFiles(oldConfig, newConfig *mcfgv1.MachineConfig) error {
 	glog.Info("Updating files")
-
-	if err := dn.writeFiles(newConfig.Spec.Config.Storage.Files); err != nil {
+	oldIgnConfig, report, err := ign.Parse(oldConfig.Spec.Config.Raw)
+	if err != nil {
+		return fmt.Errorf("failed to update files. Parsing old Ignition config failed with error: %v\nReport: %v", err, report)
+	}
+	newIgnConfig, report, err := ign.Parse(newConfig.Spec.Config.Raw)
+	if err != nil {
+		return fmt.Errorf("failed to update files. Parsing new Ignition config failed with error: %v\nReport: %v", err, report)
+	}
+	if err := dn.writeFiles(newIgnConfig.Storage.Files); err != nil {
 		return err
 	}
-	if err := dn.writeUnits(newConfig.Spec.Config.Systemd.Units); err != nil {
+	if err := dn.writeUnits(newIgnConfig.Systemd.Units); err != nil {
 		return err
 	}
-	if err := dn.deleteStaleData(oldConfig, newConfig); err != nil {
+	if err := dn.deleteStaleData(&oldIgnConfig, &newIgnConfig); err != nil {
 		return err
 	}
 	return nil
 }
 
-// deleteStaleData performs a diff of the new and the old config. It then deletes
+func restorePath(path string) error {
+	if out, err := exec.Command("cp", "-a", "--reflink=auto", origFileName(path), path).CombinedOutput(); err != nil {
+		return errors.Wrapf(err, "restoring %q from orig file %q: %s", path, origFileName(path), string(out))
+	}
+	if err := os.Remove(origFileName(path)); err != nil {
+		return errors.Wrapf(err, "deleting orig file %q: %v", origFileName(path), err)
+	}
+	return nil
+}
+
+// deleteStaleData performs a diff of the new and the old Ignition config. It then deletes
 // all the files, units that are present in the old config but not in the new one.
 // this function will error out if it fails to delete a file (with the exception
 // of simply warning if the error is ENOENT since that's the desired state).
 //nolint:gocyclo
-func (dn *Daemon) deleteStaleData(oldConfig, newConfig *mcfgv1.MachineConfig) error {
+func (dn *Daemon) deleteStaleData(oldIgnConfig, newIgnConfig *igntypes.Config) error {
 	glog.Info("Deleting stale data")
 	newFileSet := make(map[string]struct{})
-	for _, f := range newConfig.Spec.Config.Storage.Files {
+	for _, f := range newIgnConfig.Storage.Files {
 		newFileSet[f.Path] = struct{}{}
 	}
 
@@ -811,7 +873,7 @@ func (dn *Daemon) deleteStaleData(oldConfig, newConfig *mcfgv1.MachineConfig) er
 		return errors.Wrapf(err, "checking operating system")
 	}
 
-	for _, f := range oldConfig.Spec.Config.Storage.Files {
+	for _, f := range oldIgnConfig.Storage.Files {
 		if _, ok := newFileSet[f.Path]; !ok {
 			if _, err := os.Stat(noOrigFileStampName(f.Path)); err == nil {
 				if err := os.Remove(noOrigFileStampName(f.Path)); err != nil {
@@ -826,7 +888,7 @@ func (dn *Daemon) deleteStaleData(oldConfig, newConfig *mcfgv1.MachineConfig) er
 				if _, err := exec.Command("rpm", "-qf", f.Path).CombinedOutput(); err == nil {
 					// File is owned by an rpm
 					restore = true
-				} else if strings.HasPrefix(f.Path, "/etc") && (operatingSystem == machineConfigDaemonOSRHCOS) {
+				} else if strings.HasPrefix(f.Path, "/etc") && (operatingSystem == machineConfigDaemonOSRHCOS || operatingSystem == machineConfigDaemonOSFCOS) {
 					if _, err := os.Stat("/usr" + f.Path); err != nil {
 						if !os.IsNotExist(err) {
 							return err
@@ -839,11 +901,8 @@ func (dn *Daemon) deleteStaleData(oldConfig, newConfig *mcfgv1.MachineConfig) er
 				}
 
 				if restore {
-					if out, err := exec.Command("cp", "-a", "--reflink=auto", origFileName(f.Path), f.Path).CombinedOutput(); err != nil {
-						return errors.Wrapf(err, "restoring %q from orig file %q: %s", f.Path, origFileName(f.Path), string(out))
-					}
-					if err := os.Remove(origFileName(f.Path)); err != nil {
-						return errors.Wrapf(err, "deleting orig file %q: %v", origFileName(f.Path), err)
+					if err := restorePath(f.Path); err != nil {
+						return err
 					}
 					glog.V(2).Infof("Restored file %q", f.Path)
 					continue
@@ -868,7 +927,7 @@ func (dn *Daemon) deleteStaleData(oldConfig, newConfig *mcfgv1.MachineConfig) er
 
 	newUnitSet := make(map[string]struct{})
 	newDropinSet := make(map[string]struct{})
-	for _, u := range newConfig.Spec.Config.Systemd.Units {
+	for _, u := range newIgnConfig.Systemd.Units {
 		for j := range u.Dropins {
 			path := filepath.Join(pathSystemd, u.Name+".d", u.Dropins[j].Name)
 			newDropinSet[path] = struct{}{}
@@ -877,10 +936,22 @@ func (dn *Daemon) deleteStaleData(oldConfig, newConfig *mcfgv1.MachineConfig) er
 		newUnitSet[path] = struct{}{}
 	}
 
-	for _, u := range oldConfig.Spec.Config.Systemd.Units {
+	for _, u := range oldIgnConfig.Systemd.Units {
 		for j := range u.Dropins {
 			path := filepath.Join(pathSystemd, u.Name+".d", u.Dropins[j].Name)
 			if _, ok := newDropinSet[path]; !ok {
+				if _, err := os.Stat(noOrigFileStampName(path)); err == nil {
+					if err := os.Remove(noOrigFileStampName(path)); err != nil {
+						return errors.Wrapf(err, "deleting noorig file stamp %q: %v", noOrigFileStampName(path), err)
+					}
+					glog.V(2).Infof("Removing file %q completely", path)
+				} else if _, err := os.Stat(origFileName(path)); err == nil {
+					if err := restorePath(path); err != nil {
+						return err
+					}
+					glog.V(2).Infof("Restored file %q", path)
+					continue
+				}
 				glog.V(2).Infof("Deleting stale systemd dropin file: %s", path)
 				if err := os.Remove(path); err != nil {
 					newErr := fmt.Errorf("unable to delete %s: %s", path, err)
@@ -898,6 +969,18 @@ func (dn *Daemon) deleteStaleData(oldConfig, newConfig *mcfgv1.MachineConfig) er
 			if err := dn.disableUnit(u); err != nil {
 				glog.Warningf("Unable to disable %s: %s", u.Name, err)
 			}
+			if _, err := os.Stat(noOrigFileStampName(path)); err == nil {
+				if err := os.Remove(noOrigFileStampName(path)); err != nil {
+					return errors.Wrapf(err, "deleting noorig file stamp %q: %v", noOrigFileStampName(path), err)
+				}
+				glog.V(2).Infof("Removing file %q completely", path)
+			} else if _, err := os.Stat(origFileName(path)); err == nil {
+				if err := restorePath(path); err != nil {
+					return err
+				}
+				glog.V(2).Infof("Restored file %q", path)
+				continue
+			}
 			glog.V(2).Infof("Deleting stale systemd unit file: %s", path)
 			if err := os.Remove(path); err != nil {
 				newErr := fmt.Errorf("unable to delete %s: %s", path, err)
@@ -950,11 +1033,21 @@ func (dn *Daemon) disableUnit(unit igntypes.Unit) error {
 
 // writeUnits writes the systemd units to disk
 func (dn *Daemon) writeUnits(units []igntypes.Unit) error {
+	operatingSystem, err := getHostRunningOS()
+	if err != nil {
+		return errors.Wrapf(err, "checking operating system")
+	}
 	for _, u := range units {
 		// write the dropin to disk
 		for i := range u.Dropins {
 			glog.Infof("Writing systemd unit dropin %q", u.Dropins[i].Name)
 			dpath := filepath.Join(pathSystemd, u.Name+".d", u.Dropins[i].Name)
+			if _, err := os.Stat("/usr" + dpath); err == nil &&
+				(operatingSystem == machineConfigDaemonOSRHCOS || operatingSystem == machineConfigDaemonOSFCOS) {
+				if err := createOrigFile("/usr"+dpath, dpath); err != nil {
+					return err
+				}
+			}
 			if err := writeFileAtomicallyWithDefaults(dpath, []byte(u.Dropins[i].Contents)); err != nil {
 				return fmt.Errorf("failed to write systemd unit dropin %q: %v", u.Dropins[i].Name, err)
 			}
@@ -983,7 +1076,12 @@ func (dn *Daemon) writeUnits(units []igntypes.Unit) error {
 
 		if u.Contents != "" {
 			glog.Infof("Writing systemd unit %q", u.Name)
-
+			if _, err := os.Stat("/usr" + fpath); err == nil &&
+				(operatingSystem == machineConfigDaemonOSRHCOS || operatingSystem == machineConfigDaemonOSFCOS) {
+				if err := createOrigFile("/usr"+fpath, fpath); err != nil {
+					return err
+				}
+			}
 			// write the unit to disk
 			if err := writeFileAtomicallyWithDefaults(fpath, []byte(u.Contents)); err != nil {
 				return fmt.Errorf("failed to write systemd unit %q: %v", u.Name, err)
@@ -1047,7 +1145,7 @@ func (dn *Daemon) writeFiles(files []igntypes.File) error {
 				return fmt.Errorf("failed to retrieve file ownership for file %q: %v", file.Path, err)
 			}
 		}
-		if err := createOrigFile(file.Path); err != nil {
+		if err := createOrigFile(file.Path, file.Path); err != nil {
 			return err
 		}
 		if err := writeFileAtomically(file.Path, contents.Data, defaultDirectoryPermissions, mode, uid, gid); err != nil {
@@ -1080,7 +1178,7 @@ func noOrigFileStampName(fpath string) string {
 	return filepath.Join(noOrigParentDir(), fpath+".mcdnoorig")
 }
 
-func createOrigFile(fpath string) error {
+func createOrigFile(fromPath, fpath string) error {
 	if _, err := os.Stat(noOrigFileStampName(fpath)); err == nil {
 		// we already created the no orig file for this default file
 		return nil
@@ -1101,7 +1199,7 @@ func createOrigFile(fpath string) error {
 	if err := os.MkdirAll(filepath.Dir(origFileName(fpath)), 0755); err != nil {
 		return errors.Wrapf(err, "creating orig parent dir: %v", err)
 	}
-	if out, err := exec.Command("cp", "-a", "--reflink=auto", fpath, origFileName(fpath)).CombinedOutput(); err != nil {
+	if out, err := exec.Command("cp", "-a", "--reflink=auto", fromPath, origFileName(fpath)).CombinedOutput(); err != nil {
 		return errors.Wrapf(err, "creating orig file for %q: %s", fpath, string(out))
 	}
 	return nil
@@ -1161,9 +1259,12 @@ func (dn *Daemon) updateSSHKeys(newUsers []igntypes.PasswdUser) error {
 
 	// we're also appending all keys for any user to core, so for now
 	// we pass this to atomicallyWriteSSHKeys to write.
+	// we know these users are "core" ones also cause this slice went through Reconcilable
 	var concatSSHKeys string
-	for _, k := range newUsers[len(newUsers)-1].SSHAuthorizedKeys {
-		concatSSHKeys = concatSSHKeys + string(k) + "\n"
+	for _, u := range newUsers {
+		for _, k := range u.SSHAuthorizedKeys {
+			concatSSHKeys = concatSSHKeys + string(k) + "\n"
+		}
 	}
 	if !dn.mock {
 		// Note we write keys only for the core user and so this ignores the user list
@@ -1176,8 +1277,8 @@ func (dn *Daemon) updateSSHKeys(newUsers []igntypes.PasswdUser) error {
 
 // updateOS updates the system OS to the one specified in newConfig
 func (dn *Daemon) updateOS(config *mcfgv1.MachineConfig) error {
-	if dn.OperatingSystem != machineConfigDaemonOSRHCOS {
-		glog.V(2).Info("Updating of non RHCOS nodes are not supported")
+	if dn.OperatingSystem != machineConfigDaemonOSRHCOS && dn.OperatingSystem != machineConfigDaemonOSFCOS {
+		glog.V(2).Info("Updating of non-CoreOS nodes are not supported")
 		return nil
 	}
 
